"""Tool for the Google search API."""
from typing import Any

from rich.console import Console

from chatgpt_tool_hub.tools.all_tool_list import main_tool_register
from chatgpt_tool_hub.tools.base_tool import BaseTool
from chatgpt_tool_hub.tools.google_search.wrapper import GoogleSearchAPIWrapper

default_tool_name = "google-search"


class GoogleSearch(BaseTool):
    """Tool that adds the capability to query the Google search API."""

    name = default_tool_name
    description = (
        "A wrapper around Google Search. "
        "Useful for when you need to answer questions about current events. "
        "Input should be a search query."
    )
    api_wrapper: GoogleSearchAPIWrapper

    def __init__(self, console: Console = Console(), **tool_kwargs: Any):
        super().__init__(console=console)
        self.api_wrapper = GoogleSearchAPIWrapper(**tool_kwargs)

    def _run(self, query: str) -> str:
        """Use the tool."""
        return self.api_wrapper.run(query)

    async def _arun(self, query: str) -> str:
        """Use the tool asynchronously."""
        raise NotImplementedError("GoogleSearchRun does not support async")


class GoogleSearchJson(BaseTool):
    """Tool that has capability to query the Google Search API and get back json."""

    name = default_tool_name
    description = (
        "A wrapper around Google Search. "
        "Useful for when you need to answer questions about current events. "
        "Input should be a search query. Output is a JSON array of the query results"
    )
    api_wrapper: GoogleSearchAPIWrapper = None

    def __init__(self, console: Console = Console(), **tool_kwargs: Any):
        super().__init__(console=console)
        self.api_wrapper = GoogleSearchAPIWrapper(**tool_kwargs)

    def _run(self, query: str) -> str:
        """Use the tool."""
        return str(self.api_wrapper.results(query))

    async def _arun(self, query: str) -> str:
        """Use the tool asynchronously."""
        raise NotImplementedError("GoogleSearchRun does not support async")


main_tool_register.register_tool(default_tool_name,
                                 lambda console, kwargs: GoogleSearchJson(console, **kwargs),
                                 tool_input_keys=["google_api_key", "google_cse_id"])



import logging
import os

from rich.console import Console

from chatgpt_tool_hub.apps import App
from chatgpt_tool_hub.common.log import consoleHandler
from chatgpt_tool_hub.common.utils import get_from_dict_or_env
from chatgpt_tool_hub.models import build_model_params
from chatgpt_tool_hub.tools import dynamic_tool_loader


class AppFactory:
    def __init__(self, console=Console()):
        self.default_tools_list = []
        self.console = console

    def init_env(self, **kwargs):
        """ 环境初始化 """
        # set log level
        nolog_flag = get_from_dict_or_env(kwargs, "nolog", "NOLOG", "false")
        debug_flag = get_from_dict_or_env(kwargs, "debug", "DEBUG", "false")
        if str(nolog_flag).lower() in {'true', 'enable', 'yes'}:
            consoleHandler.setLevel(logging.CRITICAL)
        elif str(debug_flag).lower() in {'true', 'enable', 'yes'}:
            consoleHandler.setLevel(logging.DEBUG)
        else:
            consoleHandler.setLevel(logging.INFO)

        # default tools
        no_default_flag = get_from_dict_or_env(kwargs, "no_default", "NO_DEFAULT", "")
        if str(no_default_flag).lower() in {'true', 'enable', 'yes'}:
            self.default_tools_list = []
        else:
            self.default_tools_list = ["python", "terminal", "url-get", "meteo-weather"]

        # set proxy
        # _proxy = get_from_dict_or_env(kwargs, "proxy", "PROXY", "")
        # if not _proxy:
        #     os.environ["http_proxy"] = str(_proxy)
        #     os.environ["https_proxy"] = str(_proxy)

        # dynamic loading tool
        dynamic_tool_loader()

    def create_app(self, app_type: str = 'victorinox', tools_list: list = None, console=Console(quiet=True), **kwargs) -> App:
        tools_list = tools_list if tools_list else []

        self.init_env(**kwargs)

        if app_type == 'lite':
            from chatgpt_tool_hub.apps.lite_app import LiteApp
            app = LiteApp(**build_model_params(kwargs))
            app.create(tools_list, **kwargs)
            return app

        elif app_type == 'victorinox':
            from chatgpt_tool_hub.apps.victorinox import Victorinox

            for default_tool in self.default_tools_list:
                if default_tool not in tools_list:
                    tools_list.append(default_tool)

            # todo refactor
            if "browser" in tools_list:
                tools_list = list(filter(lambda tool: tool != "url-get", tools_list))

            app = Victorinox(console, **build_model_params(kwargs))
            app.create(tools_list, **kwargs)
            return app
        else:
            raise NotImplementedError



import logging
from typing import Any

from rich.console import Console

from chatgpt_tool_hub.common.log import LOG
from chatgpt_tool_hub.tools import BrowserTool
from chatgpt_tool_hub.tools.base_tool import BaseTool
from chatgpt_tool_hub.tools.news import news_tool_register

default_tool_name = "finance-news"


class FinanceNewsTool(BaseTool):
    
    name: str = default_tool_name
    description: str = (
        "Useful when you want to stay up-to-date on global real-time financial news."
        "The tool aggregates various types of data and information from the financial investment industry."
        "input is needless for this tool."
    )

    jin10_url: str = "https://www.jin10.com/example/jin10.com.html"

    def __init__(self, console: Console = Console(), **tool_kwargs: Any):
        super().__init__(console=console, return_direct=True)

    def _run(self, tool_input: str) -> str:
        """ finance-news 仅仅是一个 browser tool 的封装吗？ """
        return BrowserTool().run(self.jin10_url)

    async def _arun(self, tool_input: str) -> str:
        pass


news_tool_register.register_tool(default_tool_name, lambda console, kwargs: FinanceNewsTool(console, **kwargs), [])


if __name__ == "__main__":
    LOG.setLevel(logging.DEBUG)

    tool = FinanceNewsTool()
    content = tool.run("")
    print(content)



ARXIV_PROMPT = """
You are a converter from natural language to json to search research papers by arxiv api.

You should only respond in JSON format as described below.
Response Format: 
{{
    "search_query": "ti:multi+UAV" ,
    "max_results": 3,
    "sort_by": "relevance",
    "sort_order": "descending"
}}
json note:
search_query: an arXiv query string. I will teach you how to generate a search_query below.
max_results: int, range: 1~20
sort_by: The sort criterion for results: "relevance", "lastUpdatedDate", or "submittedDate"
sort_order: The sort order for results: "descending" or "ascending"

How to generate a search_query:

Each paper is divided up into a number of fields that can individually be searched. 
A field consists of field prefix, a colon, and search term.  
field prefix: ti (i.e. title), au (author), abs (abstract) and co (comment). 
A search term can be search keywords(ti, abs, co), author name(au)
All search terms must be English.

Example: 
1. ti:large+language+model
2. au:del_maestro
(Note: plus sign stands for a space in search_query)

you can use boolean operator to combine fields.
boolean operator: AND, OR, ANDNOT

Example: if we wanted all of the articles by the author Adrian DelMaestro 
with titles that did not contain the computer vision. 
A search_query should be au:del_maestro+ANDNOT+ti:computer+vision. 

When you convert, you should think step by step:
1. What the English meaning of "input" is. if no specific search terms provided, you can generate terms that you like.
2. Whose papers do I need to search for?
3. What keywords do I need to use to search for papers?
4. How can I search more accurately?
5. build the json

Input: {input}
json:"""


"""Tool for the SearxNG search API."""
from typing import Any

from pydantic import Extra
from rich.console import Console

from chatgpt_tool_hub.tools.all_tool_list import main_tool_register
from chatgpt_tool_hub.tools.base_tool import BaseTool
from chatgpt_tool_hub.tools.searxng_search.wrapper import SearxSearchWrapper

default_tool_name = "searxng-search"


class SearxSearchTool(BaseTool):
    """Tool that adds the capability to query a Searx instance."""

    name = default_tool_name
    description = (
        "A meta search engine."
        "Useful for when you need to answer questions about current events."
        "Input should be a search query."
    )
    api_wrapper: SearxSearchWrapper

    def __init__(self, console: Console = Console(), **tool_kwargs: Any):
        super().__init__(console=console)

    def _run(self, query: str) -> str:
        """Use the tool."""
        return self.api_wrapper.run(query)

    async def _arun(self, query: str) -> str:
        """Use the tool asynchronously."""
        return await self.api_wrapper.arun(query)


class SearxSearchJsonTool(BaseTool):
    """(Json result): Tool that has capability to query a Searx instance and get back json."""

    name = "searxng-search"
    description = (
        "A meta search engine."
        "Useful for when you need to answer questions about current events."
        "Input should be a search query. Output is a JSON array of the query results"
    )
    api_wrapper: SearxSearchWrapper
    num_results: int = 4

    def __init__(self, console: Console = Console(), **tool_kwargs):
        super().__init__(console=console, return_direct=True)
        self.api_wrapper = SearxSearchWrapper(**tool_kwargs)

    class Config:
        """Pydantic config."""

        extra = Extra.ignore

    def _run(self, query: str) -> str:
        """Use the tool."""
        return str(self.api_wrapper.results(query, self.num_results))

    async def _arun(self, query: str) -> str:
        """Use the tool asynchronously."""
        return (await self.api_wrapper.aresults(query, self.num_results)).__str__()


main_tool_register.register_tool(default_tool_name,
                                 lambda console, kwargs: SearxSearchTool(console, **kwargs),
                                 tool_input_keys=["searx_search_host"])



import json
from typing import Any

from rich.console import Console

from chatgpt_tool_hub.chains import LLMChain
from chatgpt_tool_hub.common.utils import get_from_dict_or_env
from chatgpt_tool_hub.models import build_model_params
from chatgpt_tool_hub.models.model_factory import ModelFactory
from chatgpt_tool_hub.prompts import PromptTemplate
from chatgpt_tool_hub.tools.base_tool import BaseTool
from chatgpt_tool_hub.tools.news import news_tool_register
from chatgpt_tool_hub.tools.news.morning_news.prompt import SUMMARY_DOCS
from chatgpt_tool_hub.tools.web_requests.get import RequestsWrapper

default_tool_name = "morning-news"


class MorningNewsTool(BaseTool):
    name: str = default_tool_name
    description: str = (
        "Use this tool when you want to get information about Daily 60 seconds morning news today. "
        "no input."
    )
    bot: Any = None
    morning_news_api_key: str = ""
    morning_news_use_llm: bool = False

    def __init__(self, console: Console = Console(), **tool_kwargs: Any):
        # 这个工具直接返回内容
        super().__init__(console=console, return_direct=True)

        self.morning_news_api_key = get_from_dict_or_env(tool_kwargs, "morning_news_api_key", "MORNING_NEWS_API_KEY")
        self.morning_news_use_llm = get_from_dict_or_env(tool_kwargs, "morning_news_use_llm", "MORNING_NEWS_USE_LLM", False)

        llm = ModelFactory().create_llm_model(**build_model_params(tool_kwargs))
        prompt = PromptTemplate(
            input_variables=["morning_news"],
            template=SUMMARY_DOCS,
        )
        self.bot = LLMChain(llm=llm, prompt=prompt)

    def _run(self, query: str) -> str:
        """Use the tool."""
        if not query:
            return "the input of tool is empty"

        morning_news_url = f"https://v2.alapi.cn/api/zaobao?token={self.morning_news_api_key}&format=json"
        _response = RequestsWrapper().get(morning_news_url)
        _response_json = json.loads(_response)
        if self.morning_news_use_llm:
            return self.bot.run(_response)
        elif _response_json.get("code") == 200 or _response_json.get("msg") == "success":
            # 不使用llm，表明人类具有先天的优越性
            _return_data = _response_json.get('data', {})
            _date = _return_data.get('date')
            _news_content = "\n".join(_return_data.get("news"))
            _weiyu = _return_data.get('weiyu')
            _image_url = _return_data.get('image')
            return f"\n今日日期：{_date}\n\n今日早报：{_news_content}\n\n今日微语：{_weiyu}\n\nURL: {_image_url}"
        else:
            return f"[{default_tool_name}] api error, error_info: {_response_json.get('msg')}"

    async def _arun(self, query: str) -> str:
        """Use the tool asynchronously."""
        raise NotImplementedError("NewsTool does not support async")


news_tool_register.register_tool(default_tool_name, lambda console, kwargs: MorningNewsTool(console, **kwargs), ["morning_news_api_key"])


if __name__ == "__main__":
    tool = MorningNewsTool(morning_news_api_key="", morning_news_use_llm=False)
    content = tool.run("给我发一下早报？")
    print(content)




import os
from typing import Optional

from chatgpt_tool_hub.common.cache import BaseCache
from chatgpt_tool_hub.common.utils import get_from_dict_or_env

verbose: bool = False
llm_cache: Optional[BaseCache] = None

# default for gpt-3.5
ALL_MAX_TOKENS_NUM = 4000

# token manage strategy
BOT_PROMPT = 1000
MEMORY_MAX_TOKENS_NUM = 600
BOT_SCRATCHPAD_MAX_TOKENS_NUM = ALL_MAX_TOKENS_NUM - BOT_PROMPT - MEMORY_MAX_TOKENS_NUM - 200


def change_memory_max_tokens(memory_max_tokens_num: int):
    global MEMORY_MAX_TOKENS_NUM, BOT_SCRATCHPAD_MAX_TOKENS_NUM
    MEMORY_MAX_TOKENS_NUM = memory_max_tokens_num
    BOT_SCRATCHPAD_MAX_TOKENS_NUM = ALL_MAX_TOKENS_NUM - BOT_PROMPT - MEMORY_MAX_TOKENS_NUM - 200


def build_model_params(kwargs: dict) -> dict:
    _api_key = get_from_dict_or_env(kwargs, "openai_api_key", "OPENAI_API_KEY")
    _proxy = get_from_dict_or_env(kwargs, "proxy", "PROXY", "")
    _model = get_from_dict_or_env(kwargs, "model_name", "MODEL_NAME", "gpt-3.5-turbo")
    _timeout = get_from_dict_or_env(kwargs, "request_timeout", "REQUEST_TIMEOUT", 120)
    # tool llm need them
    os.environ["OPENAI_API_KEY"] = str(_api_key)
    os.environ["PROXY"] = str(_proxy)
    os.environ["MODEL_NAME"] = str(_model)
    os.environ["REQUEST_TIMEOUT"] = str(_timeout)
    return {
        "temperature": get_from_dict_or_env(kwargs, "temperature", "TEMPERATURE", 0),
        "openai_api_key": _api_key,
        "proxy": _proxy,
        "model_name": _model,  # 对话模型的名称
        "top_p": 1,
        "frequency_penalty": 0.0,  # [-2,2]之间，该值越大则更倾向于产生不同的内容
        "presence_penalty": 0.0,  # [-2,2]之间，该值越大则更倾向于产生不同的内容
        "request_timeout": _timeout,
        "max_retries": 2
    }


from chatgpt_tool_hub.models.base import BaseLLM, LLM
from chatgpt_tool_hub.models.chatgpt.chatgpt import ChatOpenAI
from chatgpt_tool_hub.models.openai import OpenAI, AzureOpenAI


__all__ = [
    "ALL_MAX_TOKENS_NUM",
    "MEMORY_MAX_TOKENS_NUM",
    "BOT_SCRATCHPAD_MAX_TOKENS_NUM",

    "BaseLLM",
    "LLM",
    "OpenAI",
    "AzureOpenAI",
    "ChatOpenAI"
]


from typing import List

from rich.console import Console

from chatgpt_tool_hub.apps import App
from chatgpt_tool_hub.apps import AppFactory
from chatgpt_tool_hub.common.log import LOG
from chatgpt_tool_hub.common.utils import get_from_dict_or_env
from chatgpt_tool_hub.database import ConversationTokenBufferMemory
from chatgpt_tool_hub.engine.initialize import init_tool_engine as init_engine
from chatgpt_tool_hub.models import MEMORY_MAX_TOKENS_NUM
from chatgpt_tool_hub.models.model_factory import ModelFactory
from chatgpt_tool_hub.tools.all_tool_list import main_tool_register
from chatgpt_tool_hub.tools.base_tool import BaseTool
from chatgpt_tool_hub.tools.load_tools import load_tools


class Victorinox(App):
    def __init__(self, console=Console(), **app_kwargs):
        super().__init__()
        if not self.init_flag:
            self.llm = ModelFactory().create_llm_model(**app_kwargs)

            self.memory = ConversationTokenBufferMemory(llm=self.llm, memory_key="chat_history",
                                                        output_key='output', max_token_limit=MEMORY_MAX_TOKENS_NUM)
            self.bot_type = "chat-bot"

            self.think_depth = get_from_dict_or_env(app_kwargs, "think_depth", "THINK_DEPTH", 3)

            self.console = console

            # todo don't remove it
            self.init_flag = True

    def create(self, tools_list: list, **tools_kwargs):
        if tools_list is None:
            tools_list = []
        if not self._check_mandatory_tools(tools_list):
            raise ValueError("check mandatory tools failed")
        if self.tools or self.tools_kwargs:
            LOG.warning("refresh the config of tools")

        self.update_tool_args(tools_list, **tools_kwargs)
        self.load_tools_into_bot(**tools_kwargs)

    def add_tool(self, tool: str, **tools_kwargs):
        if not tool:
            LOG.info("no tool to add")
            return

        self.update_tool_args([tool], **tools_kwargs)
        self.load_tools_into_bot()

    def del_tool(self, tool: str, **tools_kwargs):
        if not tool:
            LOG.info("no tool to add")
            return

        self.update_tool_args([tool], is_del=True, **tools_kwargs)
        self.load_tools_into_bot()

    def update_tool_args(self, tools_list: list, is_del: bool = False, **tools_kwargs):
        if not tools_list:
            return

        if is_del:
            for tool in tools_list:
                self.tools.remove(tool)
            return

        for tool in tools_list:
            self.tools.add(tool)

        for tool_key in tools_kwargs:
            self.tools_kwargs[tool_key] = tools_kwargs[tool_key]

    def load_tools_into_bot(self, **tools_kwargs):
        try:
            tools = load_tools(list(self.tools), main_tool_register, console=self.console, **self.tools_kwargs)
        except ValueError as e:
            LOG.error(repr(e))
            raise RuntimeError("loading tool failed")

        # tool可能注册失败
        self.tools = {tool.name for tool in tools}
        LOG.info(f"use_tools={list(self.tools)}, params: {str(self.tools_kwargs)}")

        self.init_tool_engine(tools, **tools_kwargs)

    def init_tool_engine(self, tools: List[BaseTool] = list, **bot_kwargs):
        # create bots
        # todo fix verbose params
        self.engine = init_engine(tools, self.llm, bot=self.bot_type, verbose=True, memory=self.memory,
                                  max_iterations=self.think_depth, early_stopping_method="generate",
                                  console=self.console, bot_kwargs=bot_kwargs)

    def ask(self, query: str, chat_history: list = None, retry_num: int = 0) -> str:
        if self.engine is None:
            LOG.error("before calling the ask method, you should use create bot firstly")
            raise RuntimeError("app初始化失败")

        if not query:
            LOG.warning("[APP]: query is zero value")
            raise ValueError("请求为空")

        # 更新session
        if chat_history is not None:
            self._refresh_memory(chat_history)

        try:
            LOG.info(f"提问: {query}")
            return self.engine.run(query)
        except Exception as e:
            LOG.error(f"[APP] catch a Exception: {str(e)}")
            if retry_num < 1:
                return self.ask(query, chat_history, retry_num + 1)
            LOG.error("exceed retry_num")
            raise TimeoutError("超过重试次数")

    def _refresh_memory(self, chat_history: list):
        self.memory.chat_memory.clear()
        inputs = {
            "input": ""
        }
        outputs = {
            "output": ""
        }

        for item in chat_history:
            if item.get('role') == 'user':
                inputs["input"] = item.get('content')
            elif item.get('role') == 'assistant':
                outputs["output"] = item.get('content')
                self.memory.save_context(inputs, outputs)

        LOG.debug(f"Now memory: {repr(self.memory.chat_memory.messages)}")


if __name__ == "__main__":
    bot = AppFactory().create_app(tools_list=["wikipedia"])
    content = bot.ask("")
    print(content)



""" tool-hub核心prompt，改动会影响所有tool效果 """
PREFIX = """You are a helpful AI operating system called LLM-OS. 
A user named {human_prefix} is currently interacting with you.

There are three individuals present here: you, me, and the user.
Your task is to construct a JSON to assist me in resolving user issues using tools.
You only need to response me the JSON, and I will invoke the tool described and return the result to you. 
Then, you will need to think step by step and determine what is the next tool to be used and what is the input of tool.

If nothing todo or want to ask user for guidance, You need to construct a JSON that uses the 'answer-user' tool and response it to me.
The user cannot see our interaction, so you need to act on my behalf and answer the questions posed by the user.
You need to ensure that the final tool used is `answer-user`.

LLM-OS has access to the following tools:
TOOLS:
------"""

# prompt below is inspired by auto-gpt
FORMAT_INSTRUCTIONS = """You should only respond in JSON format as described below 
Response Format: 

{{{{
    "thoughts": {{{{
        "text": "thought",
        "reasoning": "reasoning",
        "criticism": "constructive self-criticism",
        "speak": "thoughts summary to say to {human_prefix}",
    }}}},
    "tool": {{{{
        "name": "the tool to use, You must use one of the tools from the list: [{tool_names}, answer-user]", 
        "input": "the input to the tool" 
    }}}}
}}}}

You should think the content of `User input` and `Your scratchpad` step by step, and then generate your thought, reasoning and self-criticism  
The strings corresponding to "text", "reasoning", "criticism", and "speak" in JSON should be described in Chinese.

Ensure the response can be parsed by Python json.loads
"""

SUFFIX = """Begin!

Previous conversation history:
{chat_history}

User input: {input}

Your scratchpad: {bot_scratchpad}

Response: """



import asyncio
import os
from typing import Any, List

from rich.console import Console
from rich.panel import Panel

from chatgpt_tool_hub.chains.llm import LLMChain
from chatgpt_tool_hub.common.calculate_token import count_string_tokens as get_token_num
from chatgpt_tool_hub.common.log import LOG
from chatgpt_tool_hub.common.utils import get_from_dict_or_env
from chatgpt_tool_hub.models import build_model_params
from chatgpt_tool_hub.models.model_factory import ModelFactory
from chatgpt_tool_hub.tools.all_tool_list import main_tool_register
from chatgpt_tool_hub.tools.base_tool import BaseTool
from chatgpt_tool_hub.tools.summary import MAP_PROMPT, REDUCE_PROMPT


class TextClipper:

    def __init__(self, max_segment_length=2500, seperator="\n"):
        self.max_segment_length = max_segment_length
        self.seperator = seperator

    def _clip_single_long_text(self, text: str) -> List[str]:
        _token_num = get_token_num(text)
        if _token_num <= self.max_segment_length:
            return [text]

        _cut_step = min(int(len(text) * self.max_segment_length / _token_num), self.max_segment_length)
        _cut_index = list(range(0, len(text), _cut_step))
        # add index to stop
        _cut_index.append(len(text))

        _iter = 0
        clip_list = []
        while _iter < len(_cut_index) - 1:
            # todo: some words or sentences may be split in two parts
            _clip_text = text[_cut_index[_iter]:_cut_index[_iter + 1]]
            _cut_token_num = get_token_num(_clip_text)
            if _cut_token_num <= self.max_segment_length:
                clip_list.append(_clip_text)
            _iter += 1
        return clip_list

    def clip(self, text: str, capacity: int = 0) -> List[str]:
        if get_token_num(text) <= self.max_segment_length:
            return [text]
        capacity = max(capacity, 0)
        segments = text.split(self.seperator)
        segments = list(filter(lambda x: x.strip() != "", segments))
        _segment_num = len(segments)

        clip_list = []
        segment_cache = ""

        now_ctn, total_ctn = 0, 0
        for iter in range(_segment_num):
            _mix_text = self.seperator.join([segment_cache, segments[iter]])
            _token_num_of_mix = get_token_num(_mix_text)
            if _token_num_of_mix > self.max_segment_length:
                # 长文本
                if now_ctn == 1:
                    clip_list.extend(self._clip_single_long_text(segment_cache))
                elif now_ctn > 1:
                    clip_list.append(segment_cache)
                total_ctn += now_ctn
                # not precise here
                if capacity != 0 and total_ctn >= capacity:
                    # early stop
                    segment_cache = ""
                    break
                segment_cache = segments[iter]
                now_ctn = 1
            else:
                segment_cache = _mix_text
                now_ctn += 1
        if segment_cache:
            clip_list.extend(self._clip_single_long_text(segment_cache))
        return clip_list


default_tool_name = "summary"


class SummaryTool(BaseTool):
    name = default_tool_name
    description = (
        "Useful when you want to summarize the content of a file. "
        "The input is a comma-separated of file_path and a number. "
        "The number represents summarizing only the first N lines of the file, "
        "and when the number is 0, it means summarizing the entire file. "
        "If you don't know what the file path is, you cannot use this tool."
    )

    message_num: int = 100
    max_segment_length: int = 2500

    map_bot: Any = None
    reduce_bot: Any = None

    def __init__(self, console: Console = Console(), **tool_kwargs: Any):
        super().__init__(console=console)
        tool_kwargs = {} if tool_kwargs is None else tool_kwargs
        self.message_num = get_from_dict_or_env(tool_kwargs, "message_num", "MESSAGE_NUM", 100)
        self.max_segment_length = get_from_dict_or_env(tool_kwargs, "max_segment_length", "MAX_SEGMENT_LENGTH", 2500)
        llm = ModelFactory().create_llm_model(**build_model_params(tool_kwargs))

        self.map_bot = LLMChain(llm=llm, prompt=MAP_PROMPT)
        self.reduce_bot = LLMChain(llm=llm, prompt=REDUCE_PROMPT)

    async def _acall(self, bot: LLMChain, clip_text_list: List[str]):
        tasks = []
        for text in clip_text_list:
            task = asyncio.create_task(bot.arun(text=text))
            tasks.append(task)
        return await asyncio.gather(*tasks)

    def _run(self, tool_input: str) -> str:
        """run the tool"""
        try:
            file_path, message_num = tool_input.split(",", 1)
        except Exception as e:
            LOG.error("[summary] failing in parsing the input of SummaryTool, "
                      "you should use a comma to split file_path and message_num"
                      f"input: {tool_input}, error: {repr(e)}")
            file_path = tool_input
            message_num = self.message_num

        if not os.path.isfile(file_path.split(",")[0]):
            return f"no file in path: {file_path}, please double-check you file path is valid"

        try:
            self.message_num = int(message_num)
        except Exception as e:
            LOG.error(repr(e))

        _clipper = TextClipper(self.max_segment_length)
        with open(file_path, "r") as f:
            source_text = f.read()

        _text = source_text
        ctn = 0
        while get_token_num(_text) >= self.max_segment_length:
            ctn += 1
            if ctn > 8:
                LOG.warning(f"[summary] 我已经map-reduce {ctn}轮了，但是文本量还是很长，避免死循环我帮你关掉了~")
            # map
            _clip_text_list = _clipper.clip(_text, self.message_num)
            # async call llm to get summary of each clip_text
            map_text_list = asyncio.run(self._acall(self.map_bot, _clip_text_list))
            map_text = _clipper.seperator.join(map_text_list)

            LOG.debug(f"[summary] round:{ctn}, map_list: {map_text}")
            # reduce
            _clip_summary_list = _clipper.clip(map_text, self.message_num)

            reduce_text_list = asyncio.run(self._acall(self.reduce_bot, _clip_summary_list))
            reduce_text = _clipper.seperator.join(reduce_text_list)
            LOG.debug(f"[summary] round:{ctn}, reduce_list: {reduce_text}")
            _text = reduce_text

            self.console.print(Panel(f"{_text}",
                                     title=f"[bright_magenta]Summary tool[/] 第{ctn}轮总结",
                                     highlight=True))
            
        if ctn > 2:
            self.console.print(Panel(f"{_text}",
                                     title=f"[bright_magenta]Summary tool[/] 最终总结",
                                     highlight=True))
            
        return _text

    async def _arun(self, file_path: str) -> str:
        """use this tool async."""
        raise NotImplementedError("summary tool not support async yet")


main_tool_register.register_tool(default_tool_name, lambda console, kwargs: SummaryTool(console, **kwargs), [])

if __name__ == "__main__":
    tool = SummaryTool(**{})
    content = tool.run("/Users/goldfish/Desktop/finance.log, 1000")
    print(content)



import json
import re
from typing import Any, Dict, Optional, Union

from chatgpt_tool_hub.common.log import LOG


def fix_and_parse_json(
    json_str: str, try_to_fix_with_gpt: bool = True
) -> Union[str, Dict[Any, Any]]:
    """Fix and parse JSON string"""
    try:
        json_str = json_str.replace("\t", "")
        return json.loads(json_str)
    except json.JSONDecodeError as _:  # noqa: F841
        try:
            json_str = correct_json(json_str)
            return json.loads(json_str)
        except json.JSONDecodeError as _:  # noqa: F841
            pass
    # Let's do something manually:
    # sometimes GPT responds with something BEFORE the braces:
    # "I'm sorry, I don't understand. Please try again."
    # {"text": "I'm sorry, I don't understand. Please try again.",
    #  "confidence": 0.0}
    # So let's try to find the first brace and then parse the rest
    #  of the string
    try:
        brace_index = json_str.index("{")
        json_str = json_str[brace_index:]
        last_brace_index = json_str.rindex("}")
        json_str = json_str[: last_brace_index + 1]
        return json.loads(json_str)
    except (json.JSONDecodeError, ValueError) as e:  # noqa: F841
        if not try_to_fix_with_gpt:
            raise e
        LOG.warn(
            "Warning: Failed to parse AI output, attempting to fix."
            "\n If you see this warning frequently, it's likely that"
            " your prompt is confusing the AI. Try changing it up"
            " slightly."
        )
        # This allows the AI to react to the error message,
        #   which usually results in it correcting its ways.
        LOG.error("Failed to fix AI output, telling the AI.")
        return json_str


def fix_json_by_finding_outermost_brackets(json_string):
    LOG.info("Attempting to fix JSON by finding outermost brackets\n")

    try:
        # Use regex to search for JSON objects
        import regex

        json_pattern = regex.compile(r"\{(?:[^{}]|(?R))*\}")
        if json_match := json_pattern.search(json_string):
            # Extract the valid JSON object from the string
            json_string = json_match.group(0)
            LOG.info("Apparently json was fixed.")
        else:
            raise ValueError("No valid JSON object found")

    except (json.JSONDecodeError, ValueError) as e:
        LOG.debug("Error: Invalid JSON: %s\n", json_string)
        json_string = {}

    return json_string


def extract_char_position(error_message: str) -> int:
    """Extract the character position from the JSONDecodeError message.

    Args:
        error_message (str): The error message from the JSONDecodeError
          exception.

    Returns:
        int: The character position.
    """
    import re

    char_pattern = re.compile(r"\(char (\d+)\)")
    if match := char_pattern.search(error_message):
        return int(match[1])
    else:
        raise ValueError("Character position not found in the error message.")


def add_quotes_to_property_names(json_string: str) -> str:
    """
    Add quotes to property names in a JSON string.

    Args:
        json_string (str): The JSON string.

    Returns:
        str: The JSON string with quotes added to property names.
    """

    def replace_func(match):
        return f'"{match.group(1)}":'

    property_name_pattern = re.compile(r"(\w+):")
    corrected_json_string = property_name_pattern.sub(replace_func, json_string)

    try:
        json.loads(corrected_json_string)
        return corrected_json_string
    except json.JSONDecodeError as e:
        raise e


def balance_braces(json_string: str) -> Optional[str]:
    """
    Balance the braces in a JSON string.

    Args:
        json_string (str): The JSON string.

    Returns:
        str: The JSON string with braces balanced.
    """

    open_braces_count = json_string.count("{")
    close_braces_count = json_string.count("}")

    while open_braces_count > close_braces_count:
        json_string += "}"
        close_braces_count += 1

    while close_braces_count > open_braces_count:
        json_string = json_string.rstrip("}")
        close_braces_count -= 1

    try:
        json.loads(json_string)
        return json_string
    except json.JSONDecodeError:
        pass


def fix_invalid_escape(json_str: str, error_message: str) -> str:
    while error_message.startswith("Invalid \\escape"):
        bad_escape_location = extract_char_position(error_message)
        json_str = json_str[:bad_escape_location] + json_str[bad_escape_location + 1 :]
        try:
            json.loads(json_str)
            return json_str
        except json.JSONDecodeError as e:
            LOG.info(f"json loads error - fix invalid escape {repr(e)}")
            error_message = str(e)
    return json_str


def correct_json(json_str: str) -> str:
    """
    Correct common JSON errors.

    Args:
        json_str (str): The JSON string.
    """

    try:
        LOG.info(f"json: {repr(json_str)}")
        json.loads(json_str)
        return json_str
    except json.JSONDecodeError as e:
        LOG.info(f"json loads error: {repr(e)}")
        error_message = str(e)
        if error_message.startswith("Invalid \\escape"):
            json_str = fix_invalid_escape(json_str, error_message)
        if error_message.startswith("Expecting property name enclosed in double quotes"):
            json_str = add_quotes_to_property_names(json_str)
            try:
                json.loads(json_str)
                return json_str
            except json.JSONDecodeError as e:
                LOG.info(f"json loads error - add quotes: {repr(e)}")
                error_message = str(e)
        if balanced_str := balance_braces(json_str):
            return balanced_str
    return json_str



from chatgpt_tool_hub.prompts import PromptTemplate

PROMPT = """
你是一位专业的研究分析师。我会给你提供某个长文档的一部分，你需要使用你的编辑和写作技能为它生成一个总结。
你要尽可能保留内容细节，比如时间、地点、人物，同时在摘要中保持上下文。
你看到的这部分文档头和尾可能不完整，你需要用少量文字提醒，来保证未来合并总结时能根据你这部分的总结还原出完整内容

要求：
1. 用中文总结
2. 不能超过800个字

Input: {text} 

Output:
"""

MAP_PROMPT = PromptTemplate(
    input_variables=["text"],
    template=PROMPT,
)



"""Chain that takes in an input and produces an action and action input."""
from __future__ import annotations

import json
import os
import tempfile
from abc import abstractmethod
from pathlib import Path
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

import yaml
from pydantic import BaseModel, root_validator, validator
from rich.console import Console

from chatgpt_tool_hub.chains import LLMChain
from chatgpt_tool_hub.common.calculate_token import count_string_tokens
from chatgpt_tool_hub.common.callbacks import BaseCallbackManager
from chatgpt_tool_hub.common.log import LOG
from chatgpt_tool_hub.common.schema import BotAction, BotFinish, BaseMessage
from chatgpt_tool_hub.models import ALL_MAX_TOKENS_NUM, BOT_SCRATCHPAD_MAX_TOKENS_NUM
from chatgpt_tool_hub.models.base import BaseLLM
from chatgpt_tool_hub.prompts import BasePromptTemplate
from chatgpt_tool_hub.prompts import PromptTemplate
from chatgpt_tool_hub.tools import SummaryTool
from chatgpt_tool_hub.tools.base_tool import BaseTool


class Bot(BaseModel):
    """Class responsible for calling the language model and deciding the action.

    This is driven by an LLMChain. The prompt in the LLMChain MUST include
    a variable called "bot_scratchpad" where the bot can put its
    intermediary work.
    """

    llm_chain: LLMChain
    allowed_tools: Optional[List[str]] = None
    return_values: List[str] = ["output"]
    # 当bot未按要求回复时，最多重试次数
    max_parse_retry_num: int = 1
    max_token_num: int = ALL_MAX_TOKENS_NUM

    console: Console = None

    class Config:
        """Configuration for this pydantic object."""

        arbitrary_types_allowed = True

    @validator("console")
    def set_console(cls, console: Console) -> Console:
        return console or Console()

    @abstractmethod
    def _extract_tool_and_input(self, text: str) -> Optional[Tuple[str, str]]:
        """Extract tool and tool input from llm output."""

    def _fix_text(self, text: str) -> str:
        """Fix the text."""
        raise ValueError("fix_text not implemented for this bot.")

    @property
    def _stop(self) -> List[str]:
        return [
            f"\n{self.observation_prefix.rstrip()}",
            f"\n\t{self.observation_prefix.rstrip()}",
        ]

    def _construct_scratchpad(
        self, intermediate_steps: List[Tuple[BotAction, str]]
    ) -> Union[str, List[BaseMessage]]:
        """Construct the scratchpad that lets the bot continue its thought process."""
        thoughts = ""
        for action, observation in intermediate_steps:
            thoughts += f"The JSON you responded me: \n{action.log}\n\n"
            thoughts += f"The content returned after invoking the tool: \n{observation}\n\n"
        return thoughts

    def _crop_full_input(self, inputs: str) -> str:
        """ crop too long text """
        if not inputs:
            return inputs
        _input = inputs

        while count_string_tokens(_input) >= BOT_SCRATCHPAD_MAX_TOKENS_NUM:

            # compress text size
            temp_file = tempfile.mkstemp()
            file_path = temp_file[1]

            with open(file_path, "w") as f:
                f.write(_input + "\n")
            # 总结
            _input = SummaryTool(max_segment_length=2000).run(f"{str(file_path)}, 0")
            try:
                os.remove(file_path)
            except Exception as e:
                LOG.debug(f"remove {file_path} failed... error_info: {repr(e)}")

        return _input

    def _get_next_action(self, full_inputs: Dict[str, str]) -> BotAction:
        full_output = self.llm_chain.predict(**full_inputs)
        parsed_output = self._extract_tool_and_input(full_output)

        action_input = "None" if parsed_output is None else parsed_output[1]
        LOG.info(f"输入: {action_input}")
        retry_num = 0
        while parsed_output is None:
            retry_num += 1
            if retry_num > self.max_parse_retry_num:
                raise ValueError(f"Could not parse LLM output: `{parsed_output}`")

            full_output = self._fix_text(full_output)
            full_inputs["bot_scratchpad"] += full_output
            output = self.llm_chain.predict(**full_inputs)
            # LOG.debug("(fix_text): retry response: " + str(output))
            parsed_output = self._extract_tool_and_input(output)
        return BotAction(
            tool=parsed_output[0], tool_input=parsed_output[1], log=full_output
        )

    def plan(
        self, intermediate_steps: List[Tuple[BotAction, str]], **kwargs: Any
    ) -> Union[BotAction, BotFinish]:
        """Given input, decided what to do.

        Args:
            intermediate_steps: Steps the LLM has taken to date,
                along with observations
            **kwargs: User inputs.

        Returns:
            Action specifying what tool to use.
        """
        full_inputs = self.get_full_inputs(intermediate_steps, **kwargs)
        action = self._get_next_action(full_inputs)
        if action.tool == self.finish_tool_name:
            return BotFinish({"output": action.tool_input}, action.log)
        return action

    def get_full_inputs(
        self, intermediate_steps: List[Tuple[BotAction, str]], **kwargs: Any
    ) -> Dict[str, Any]:
        """Create the full inputs for the LLMChain from intermediate steps."""
        thoughts = self._construct_scratchpad(intermediate_steps)
        new_inputs = {"bot_scratchpad": self._crop_full_input(thoughts), "stop": self._stop}
        return {**kwargs, **new_inputs}

    def prepare_for_new_call(self) -> None:
        """Prepare the bot for new call, if needed."""
        pass

    @property
    def finish_tool_name(self) -> str:
        """Name of the tool to use to finish the chain."""
        return "Final Answer"

    @property
    def input_keys(self) -> List[str]:
        """Return the input keys.

        :meta private:
        """
        return list(set(self.llm_chain.input_keys) - {"bot_scratchpad"})

    @root_validator(allow_reuse=True)
    def validate_prompt(cls, values: Dict) -> Dict:
        """Validate that prompt matches format."""
        prompt = values["llm_chain"].prompt
        if "bot_scratchpad" not in prompt.input_variables:
            LOG.warning(
                "`bot_scratchpad` should be a variable in prompt.input_variables."
                " Did not find it, so adding it at the end."
            )
            prompt.input_variables.append("bot_scratchpad")
            if isinstance(prompt, PromptTemplate):
                prompt.template += "\n{bot_scratchpad}"
            else:
                raise ValueError(f"Got unexpected prompt type {type(prompt)}")
        return values

    @property
    @abstractmethod
    def observation_prefix(self) -> str:
        """Prefix to append the observation with."""

    @property
    @abstractmethod
    def llm_prefix(self) -> str:
        """Prefix to append the LLM call with."""

    @classmethod
    @abstractmethod
    def create_prompt(cls, tools: Sequence[BaseTool]) -> BasePromptTemplate:
        """Create a prompt for this class."""

    @classmethod
    def _validate_tools(cls, tools: Sequence[BaseTool]) -> None:
        """Validate that appropriate tools are passed in."""
        pass

    @classmethod
    def from_llm_and_tools(
        cls,
        llm: BaseLLM,
        tools: Sequence[BaseTool],
        console: Console = Console(),
        callback_manager: Optional[BaseCallbackManager] = None,
        **kwargs: Any,
    ) -> Bot:
        """Construct an bot from an LLM and tools."""
        cls._validate_tools(tools)
        llm_chain = LLMChain(
            llm=llm,
            prompt=cls.create_prompt(tools),
            callback_manager=callback_manager,
        )
        tool_names = [tool.name for tool in tools]
        return cls(llm_chain=llm_chain, console=console, allowed_tools=tool_names, **kwargs)

    def return_stopped_response(
        self,
        early_stopping_method: str,
        intermediate_steps: List[Tuple[BotAction, str]],
        max_iterations: int,
        **kwargs: Any,
    ) -> BotFinish:
        """Return response when bot has been stopped due to max iterations."""
        if early_stopping_method == "force":
            # `force` just returns a constant string
            return BotFinish({"output": "Bot stopped due to max iterations."}, "")
        elif early_stopping_method == "generate":
            # Generate does one final forward pass
            thoughts = ""
            for action, observation in intermediate_steps:
                thoughts += action.log
                thoughts += (
                    f"\n{self.observation_prefix}{observation}\n{self.llm_prefix}"
                )
            # Adding to the previous steps, we now tell the LLM to make a final pred
            thoughts += (
                f"你超过了tool使用次数限制: 最多{max_iterations}次。"
                "你现在需要总结当前你了解到的所有信息，来反馈给人类。"
                "你需要生成一个final answer:"
            )
            new_inputs = {"bot_scratchpad": thoughts, "stop": self._stop}
            full_inputs = {**kwargs, **new_inputs}
            full_output = self.llm_chain.predict(**full_inputs)
            # We try to extract a final answer
            action, action_input = self._extract_tool_and_input(full_output)

            if action:
                return (
                    BotFinish({"output": action_input}, full_output)
                    if action.lower() in ['answer-user']
                    else BotFinish({"output": full_output}, full_output)
                )
            else:
                # If we cannot extract, we just return the full output
                return BotFinish({"output": full_output}, full_output)

        else:
            raise ValueError(
                "early_stopping_method should be one of `force` or `generate`, "
                f"got {early_stopping_method}"
            )

    @property
    @abstractmethod
    def _bot_type(self) -> str:
        """Return Identifier of bot type."""

    def dict(self, **kwargs: Any) -> Dict:
        """Return dictionary representation of bot."""
        _dict = super().dict()
        _dict["_type"] = self._bot_type
        return _dict

    def save(self, file_path: Union[Path, str]) -> None:
        """Save the bot.

        Args:
            file_path: Path to file to save the bot to.

        Example:
        .. code-block:: python

            # If working with bot executor
            bot.bot.save(file_path="path/bot.yaml")
        """
        # Convert file to Path object.
        save_path = Path(file_path) if isinstance(file_path, str) else file_path
        directory_path = save_path.parent
        directory_path.mkdir(parents=True, exist_ok=True)

        # Fetch dictionary to save
        bot_dict = self.dict()

        if save_path.suffix == ".json":
            with open(file_path, "w") as f:
                json.dump(bot_dict, f, indent=4)
        elif save_path.suffix == ".yaml":
            with open(file_path, "w") as f:
                yaml.dump(bot_dict, f, default_flow_style=False)
        else:
            raise ValueError(f"{save_path} must be json or yaml")

    async def aplan(
        self, intermediate_steps: List[Tuple[BotAction, str]], **kwargs: Any
    ) -> Union[BotAction, BotFinish]:
        """Given input, decided what to do.

        Args:
            intermediate_steps: Steps the LLM has taken to date,
                along with observations
            **kwargs: User inputs.

        Returns:
            Action specifying what tool to use.
        """
        full_inputs = self.get_full_inputs(intermediate_steps, **kwargs)
        action = await self._aget_next_action(full_inputs)
        if action.tool == self.finish_tool_name:
            return BotFinish({"output": action.tool_input}, action.log)
        return action

    async def _aget_next_action(self, full_inputs: Dict[str, str]) -> BotAction:
        full_output = await self.llm_chain.apredict(**full_inputs)
        parsed_output = self._extract_tool_and_input(full_output)
        retry_num = 0
        while parsed_output is None:
            retry_num += 1
            if retry_num > self.max_parse_retry_num:
                raise ValueError(f"Could not parse LLM output: `{parsed_output}`")

            full_output = self._fix_text(full_output)
            full_inputs["bot_scratchpad"] += full_output
            output = await self.llm_chain.apredict(**full_inputs)
            full_output += output
            parsed_output = self._extract_tool_and_input(full_output)
        return BotAction(
            tool=parsed_output[0], tool_input=parsed_output[1], log=full_output
        )



"""Lightweight wrapper around requests library, with async support."""
from typing import Any, Dict, Optional

import aiohttp
import requests
from pydantic import BaseModel, Extra, root_validator

from chatgpt_tool_hub.common.log import LOG
from chatgpt_tool_hub.common.utils import get_from_dict_or_env
from chatgpt_tool_hub.tools.web_requests import DEFAULT_HEADER

browser: Any = None  # ChromeWebDriver or None


class RequestsWrapper(BaseModel):
    """Lightweight wrapper around requests library."""

    headers: Optional[Dict[str, str]] = dict()
    aiosession: Optional[aiohttp.ClientSession] = None

    proxy: Optional[str]

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.ignore
        arbitrary_types_allowed = True

    @root_validator()
    def validate_environment(cls, values: Dict) -> Dict:
        """get proxy param from environment."""
        proxy = get_from_dict_or_env(
            values, 'proxy', "PROXY", ""
        )
        values["proxy"] = proxy

        return values

    def get(self, url: str, params: Dict[str, Any] = None, raise_for_status: bool = False, **kwargs) -> str:
        """GET the URL and return the text."""
        self.headers.update(DEFAULT_HEADER)
        proxies = {
            'http': self.proxy,
            'https': self.proxy,
        }
        response = requests.get(url, headers=self.headers, params=params, proxies=proxies, **kwargs)
        if raise_for_status:
            try:
                response.raise_for_status()
            except Exception as e:
                LOG.error(f"RequestsWrapper.get status_code is not good: {repr(e)}")

        return response.text

    def post(self, url: str, data: Dict[str, Any]) -> str:
        """POST to the URL and return the text."""
        self.headers = (
            self.headers.update(DEFAULT_HEADER)
            if self.headers
            else {}.update(DEFAULT_HEADER)
        )
        return requests.post(url, data=data, headers=self.headers).text

    def patch(self, url: str, data: Dict[str, Any]) -> str:
        """PATCH the URL and return the text."""
        self.headers = (
            self.headers.update(DEFAULT_HEADER)
            if self.headers
            else {}.update(DEFAULT_HEADER)
        )
        return requests.patch(url, json=data, headers=self.headers).text

    def put(self, url: str, data: Dict[str, Any]) -> str:
        """PUT the URL and return the text."""
        self.headers = (
            self.headers.update(DEFAULT_HEADER)
            if self.headers
            else {}.update(DEFAULT_HEADER)
        )
        return requests.put(url, json=data, headers=self.headers).text

    def delete(self, url: str) -> str:
        """DELETE the URL and return the text."""
        self.headers = (
            self.headers.update(DEFAULT_HEADER)
            if self.headers
            else {}.update(DEFAULT_HEADER)
        )
        return requests.delete(url, headers=self.headers).text

    async def _arequest(self, method: str, url: str, **kwargs: Any) -> str:
        """Make an async request."""
        self.headers = (
            self.headers.update(DEFAULT_HEADER)
            if self.headers
            else {}.update(DEFAULT_HEADER)
        )
        if not self.aiosession:
            async with aiohttp.ClientSession() as session:
                async with session.request(
                    method, url, headers=self.headers, **kwargs
                ) as response:
                    return await response.text()
        else:
            async with self.aiosession.request(
                method, url, headers=self.headers, **kwargs
            ) as response:
                return await response.text()

    async def aget(self, url: str) -> str:
        """GET the URL and return the text asynchronously."""
        return await self._arequest("GET", url)

    async def apost(self, url: str, data: Dict[str, Any]) -> str:
        """POST to the URL and return the text asynchronously."""
        return await self._arequest("POST", url, json=data)

    async def apatch(self, url: str, data: Dict[str, Any]) -> str:
        """PATCH the URL and return the text asynchronously."""
        return await self._arequest("PATCH", url, json=data)

    async def aput(self, url: str, data: Dict[str, Any]) -> str:
        """PUT the URL and return the text asynchronously."""
        return await self._arequest("PUT", url, json=data)

    async def adelete(self, url: str) -> str:
        """DELETE the URL and return the text asynchronously."""
        return await self._arequest("DELETE", url)



"""Load tools."""
from typing import Any, List
from typing import Optional

from chatgpt_tool_hub.common.callbacks import BaseCallbackManager
from chatgpt_tool_hub.common.log import LOG
from chatgpt_tool_hub.tools.base_tool import BaseTool
from chatgpt_tool_hub.tools.tool_register import ToolRegister
from rich.console import Console


def load_tools(
    tool_names: List[str],
    tool_register: ToolRegister,
    console: Console = Console(),
    callback_manager: Optional[BaseCallbackManager] = None,
    **kwargs: Any,
) -> List[BaseTool]:
    """Load tools based on their name.

    Args:
        tool_names: name of tools to load.
        tool_register: the registered object of tools
        console: rich console
        callback_manager: Optional callback manager. If not provided, default global callback manager will be used.
    Returns:
        List of tools.
    """
    tools = []
    all_tool_dict = tool_register.get_registered_tool()
    for name in tool_names:
        if name in all_tool_dict:
            _get_tool_func, extra_keys = all_tool_dict[name]

            # consistency validation between input and required params
            try:
                if missing_keys := set(extra_keys).difference(kwargs):
                    raise ValueError(
                        f"Tool {name} requires some parameters that were not "
                        f"provided: {missing_keys}"
                    )
            except Exception as e:
                LOG.info(f"[{name}] init failed, error_info: {repr(e)}")
                tool_register.unregister_tool(name)
                continue

            # todo double check here
            #
            try:
                # sub_kwargs = {k: kwargs[k] for k in extra_keys if k in kwargs}
                tool = _get_tool_func(console, kwargs)
            except Exception as e:
                console.log(f"[dim][bold red]{name}[/] init failed, error_info: {repr(e)}")
                tool_register.unregister_tool(name)
                continue

            if callback_manager is not None:
                tool.callback_manager = callback_manager

            tools.append(tool)
        else:
            LOG.error(f"Now registered tools are: {list(all_tool_dict)}")
            raise ValueError(f"Got unknown tool: {name}")
    return crop_tools(tools)


def crop_tools(tools: List[BaseTool]) -> List[BaseTool]:
    if len(tools) > 10:
        LOG.warning(f"loading too many tools, some tools after {tools[10].name} will be ignored...")
        return tools[:10]
    # todo: 检查description是否超出限制
    return tools



import os

from chatgpt_tool_hub.common.log import LOG


def get_packages(path):
    """获取指定路径下的所有包名"""
    packages = []
    for name in os.listdir(path):
        full_path = os.path.join(path, name)
        if os.path.isdir(full_path) and os.path.exists(os.path.join(full_path, "__init__.py")):
            packages.append(name)
    return packages


def dynamic_tool_loader():
    try:
        all_tool_package_list = get_packages(f"{os.path.dirname(os.path.abspath(__file__))}")
    except Exception as e:
        LOG.debug(f"get_packages error: {repr(e)}")
        all_tool_package_list = ["meteo", "system", "web_requests", "wikipedia"]
        LOG.debug(f"Detected main tool package: {repr(all_tool_package_list)}")

    for package_name in all_tool_package_list:
        try:
            import importlib
            importlib.import_module(f"chatgpt_tool_hub.tools.{package_name}")
        except Exception as e:
            LOG.info(f"[{package_name}] init failed, error_info: {repr(e)}")


from chatgpt_tool_hub.tools.python import PythonREPLTool
from chatgpt_tool_hub.tools.summary import SummaryTool
from chatgpt_tool_hub.tools.terminal import TerminalTool
from chatgpt_tool_hub.tools.web_requests import BrowserTool


__all__ = [
    "SummaryTool",
    "PythonREPLTool",
    "TerminalTool",
    "BrowserTool",
    "get_packages",
    "dynamic_tool_loader"
]



from typing import Any

from rich.console import Console

from chatgpt_tool_hub.tools.all_tool_list import main_tool_register
from chatgpt_tool_hub.tools.base_tool import BaseTool

default_tool_name = "answer-user"


class AnswerUserTool(BaseTool):
    """Tool for answering User."""

    name = default_tool_name
    description = (
        "To provide your final answer, or if no other tool is needed, use this tool to communicate with user."
        "The input is the words you wish to say to the user."
    )

    def __init__(self, console: Console = Console(), **tool_kwargs: Any):
        super().__init__(console=console, return_direct=True)

    def _run(self, query: str) -> str:
        """Use the AnswerUserTool tool."""
        return query

    async def _arun(self, query: str) -> str:
        """Use the AnswerUserTool asynchronously."""
        raise NotImplementedError("AnswerUserTool does not support async")


main_tool_register.register_tool(default_tool_name, lambda console, kwargs: AnswerUserTool(console, **kwargs), [])



import logging
from typing import Any, Dict, Optional

from pydantic import BaseModel, Extra, root_validator
from rich.console import Console

from chatgpt_tool_hub.common.log import LOG
from chatgpt_tool_hub.common.utils import get_from_dict_or_env
from chatgpt_tool_hub.tools.all_tool_list import main_tool_register
from chatgpt_tool_hub.tools.base_tool import BaseTool
from chatgpt_tool_hub.tools.web_requests import filter_text

default_tool_name = "browser"


browser: Any = None  # ChromeWebDriver or None


class ChromeBrowser(BaseModel):
    """Lightweight wrapper around requests library."""

    headers: Optional[Dict[str, str]] = dict()

    proxy: Optional[str]

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.ignore
        arbitrary_types_allowed = True

    @root_validator()
    def validate_environment(cls, values: Dict) -> Dict:
        """Validate that browser param exists in environment."""
        proxy = get_from_dict_or_env(
            values, 'proxy', "PROXY", ""
        )
        values["proxy"] = proxy

        try:
            from selenium import webdriver
        except ImportError:
            raise ImportError(
                "selenium is not installed. "
                "Please install it with `pip install selenium`"
            )

        try:
            global browser
            if browser is None:
                browser = cls._build_browser_option(proxy)
        except ImportError:
            raise ImportError(
                "webdriver_manager is not installed. "
                "Please install it with `pip install webdriver_manager`"
            )

        return values

    @classmethod
    def _build_browser_option(cls, proxy: str = ""):
        from selenium import webdriver
        from selenium.webdriver.chrome.options import Options

        options = Options()
        options.add_argument(
            "user-agent=Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
            "AppleWebKit/537.36 (KHTML, like Gecko) Chrome/112.0.5615.49 Safari/537.36"
        )
        options.add_argument("--headless")
        options.add_argument("--disable-gpu")
        options.add_argument("--disable-extensions")
        options.add_argument("--ignore-certificate-errors")
        options.add_argument("--no-sandbox")
        options.add_argument(f"--proxy-server={proxy}")
        # 设置代理
        from webdriver_manager.chrome import ChromeDriverManager
        return webdriver.Chrome(
            executable_path=ChromeDriverManager().install(), options=options
        )

    def get(self, url: str, **kwargs) -> str:
        """GET the URL and return the text."""
        if not browser:
            return "chrome browser was not initialized"
        from selenium.webdriver.support.wait import WebDriverWait
        from selenium.webdriver.support import expected_conditions as EC
        from selenium.webdriver.common.by import By

        # Network.setExtraHTTPHeaders command to the browser's DevTools.
        # This method allows setting multiple headers at once.
        # browser.execute_cdp_cmd("Network.setExtraHTTPHeaders", {"headers": self.headers})

        browser.get(url)

        # inspired by auto-gpt
        WebDriverWait(browser, 10).until(
            EC.presence_of_element_located((By.TAG_NAME, "body"))
        )
        # Get the HTML content directly from the browser's DOM
        _content = browser.execute_script("return document.body.outerHTML;")

        if len(browser.window_handles) > 2:
            browser.close()  # 退出多余页面, 节省内存

        return _content


class BrowserTool(BaseTool):
    """Tool for making a GET request to an API endpoint."""

    name = default_tool_name
    description = (
        "A Google Chrome browser. Use this when you need to get specific content from a website. "
        "Input should be a url (i.e. https://github.com/goldfishh/chatgpt-tool-hub). "
        "The output will be the text response of browser. "
    )

    browser: ChromeBrowser = None

    def __init__(self, console: Console = Console(), **tool_kwargs: Any):
        # 这个工具直接返回内容
        super().__init__(console=console, return_direct=False)

        self.browser = ChromeBrowser(**tool_kwargs)

    def _run(self, url: str) -> str:
        """Run the tool."""
        try:
            html = self.browser.get(url)
            _content = filter_text(html)
            LOG.debug(f"[browser] output: {str(_content)}")
        except Exception as e:
            LOG.error(f"[browser] {str(e)}")
            _content = repr(e)
        return _content

    async def _arun(self, url: str) -> str:
        """Run the tool asynchronously."""
        raise NotImplementedError("not support run this tool in async")


main_tool_register.register_tool(default_tool_name, lambda console, kwargs: BrowserTool(console, **kwargs), [])


if __name__ == "__main__":
    LOG.setLevel(logging.DEBUG)
    tool = BrowserTool()
    content = tool.run("https://github.com/goldfishh/chatgpt-tool-hub")
    print(content)



import importlib
import os

from chatgpt_tool_hub.common.log import LOG
from chatgpt_tool_hub.tools import get_packages
from chatgpt_tool_hub.tools.tool_register import ToolRegister

news_tool_register = ToolRegister()
try:
    all_tool_package_list = get_packages(f"{os.path.dirname(os.path.abspath(__file__))}")
except Exception as e:
    LOG.debug(f"get_packages error: {repr(e)}")
    all_tool_package_list = ["finance_news", "morning_news", "news_api"]
    LOG.debug(f"Detected main tool package: {repr(all_tool_package_list)}")

for package_name in all_tool_package_list:
    try:
        importlib.import_module(f"chatgpt_tool_hub.tools.news.{package_name}")
    except Exception as e:
        LOG.info(f"[news.{package_name}] init failed, error_info: {repr(e)}")


from chatgpt_tool_hub.tools.news.tool import NewsTool

__all__ = [
    "NewsTool",
    "news_tool_register"
]



import os
import tempfile
from typing import Any

from rich.console import Console

from chatgpt_tool_hub.chains import LLMChain
from chatgpt_tool_hub.common.log import LOG
from chatgpt_tool_hub.common.utils import get_from_dict_or_env
from chatgpt_tool_hub.models import build_model_params
from chatgpt_tool_hub.models.model_factory import ModelFactory
from chatgpt_tool_hub.prompts import PromptTemplate
from chatgpt_tool_hub.tools.all_tool_list import main_tool_register
from chatgpt_tool_hub.tools.arxiv_search.api_prompt import ARXIV_PROMPT
from chatgpt_tool_hub.tools.arxiv_search.wrapper import ArxivAPIWrapper
from chatgpt_tool_hub.tools.base_tool import BaseTool
from chatgpt_tool_hub.tools.summary import SummaryTool

default_tool_name = "arxiv"


class ArxivTool(BaseTool):
    """ a tool to call arxiv api """

    name = default_tool_name
    description = (
        "Useful for when you need to answer questions about scientific research or search for papers "
        "Like: which papers has a certain author published? "
        "Input should be the title or abstract keywords or author names of a paper you want to search. "
    )
    bot: Any = None

    api_wrapper: ArxivAPIWrapper = None

    def __init__(self, console: Console = Console(), **tool_kwargs: Any):
        super().__init__(console=console, return_direct=True)

        self.api_wrapper = ArxivAPIWrapper(**tool_kwargs)
        self.arxiv_summary = get_from_dict_or_env(tool_kwargs, "arxiv_summary", "ARXIV_SUMMARY", True)

        llm = ModelFactory().create_llm_model(**build_model_params(tool_kwargs))
        prompt = PromptTemplate(
            input_variables=["input"],
            template=ARXIV_PROMPT,
        )
        self.bot = LLMChain(llm=llm, prompt=prompt)

    def _run(self, query: str) -> str:
        """Use the Arxiv tool."""

        _llm_response = self.bot.run(query)
        LOG.info(f"[arxiv]: search_query: {_llm_response}")

        _api_response = self.api_wrapper.run(_llm_response)

        if not self.arxiv_summary:
            return _api_response

        temp_file = tempfile.mkstemp()
        file_path = temp_file[1]
        with open(file_path, "w") as f:
            f.write(_api_response + "\n")

        _input = SummaryTool(max_segment_length=2400).run(f"{str(file_path)}, 0")
        try:
            os.remove(file_path)
        except Exception as e:
            LOG.debug(f"remove {file_path} failed... error_info: {repr(e)}")

        return _input

    async def _arun(self, query: str) -> str:
        """Use the Arxiv tool asynchronously."""
        raise NotImplementedError("ArxivTool does not support async")


main_tool_register.register_tool(default_tool_name, lambda console, kwargs: ArxivTool(console, **kwargs), [])


if __name__ == "__main__":
    tool = ArxivTool()
    content = tool.run("帮我找找吴恩达写的论文")
    print(content)



from typing import Any

from rich.console import Console

from chatgpt_tool_hub.engine import ToolEngine
from chatgpt_tool_hub.engine.initialize import init_tool_engine
from chatgpt_tool_hub.models import build_model_params
from chatgpt_tool_hub.models.model_factory import ModelFactory
from chatgpt_tool_hub.tools.all_tool_list import main_tool_register
from chatgpt_tool_hub.tools.base_tool import BaseTool
from chatgpt_tool_hub.tools.load_tools import load_tools
from chatgpt_tool_hub.tools.news import news_tool_register

default_tool_name = "news"


class NewsTool(BaseTool):
    name: str = default_tool_name
    description: str = (
        "当你想要获取实时新闻资讯时可以使用该工具，你能获取任何与新闻有关的信息。"
        "该工具包含了金融、早报和news-api三个子工具，访问这些工具前你需要先访问本工具。"
        "工具输入：你目前了解到的所有信息的总结 和 用户想获取的新闻内容。"
    )
    engine: ToolEngine = Any

    def __init__(self, console: Console = Console(), **tool_kwargs: Any):
        # 这个工具直接返回内容
        super().__init__(console=console, return_direct=True)

        tools = load_tools(news_tool_register.get_registered_tool_names(), news_tool_register, console, **tool_kwargs)
        llm = ModelFactory().create_llm_model(**build_model_params(tool_kwargs))

        # subtool 思考深度暂时固定为2
        self.engine = init_tool_engine(tools, llm, bot="qa-bot", verbose=True, console=self.console,
                                       max_iterations=2, early_stopping_method="force")

    def _run(self, query: str) -> str:
        """Use the tool."""
        if not query:
            return "the input of tool is empty"
        return (
            self.engine.run(query)
            if self.engine
            else "the tool was not initialized"
        )

    async def _arun(self, query: str) -> str:
        """Use the tool asynchronously."""
        raise NotImplementedError("NewsTool does not support async")

    def get_tool_list(self):
        return news_tool_register.get_registered_tool_names()


main_tool_register.register_tool(default_tool_name, lambda console, kwargs: NewsTool(console, **kwargs), [])




from chatgpt_tool_hub.prompts.prompt import PromptTemplate

API_URL_PROMPT_TEMPLATE = """你收到一个的调用API文档: {api_docs}
你的任务是使用这个文档，生成完整的API URL以回答用户的问题

生成URL要求如下：
1. 你必须按文档给出的BASE URL和endpoint构造URL，不能随意编造
2. 构建的API URL要保证能获得回答问题所需的必要信息
3. 请注意，有意排除API调用中不必要的数据

问题: {question}

API url:"""

API_URL_PROMPT = PromptTemplate(
    input_variables=[
        "api_docs",
        "question",
    ],
    template=API_URL_PROMPT_TEMPLATE,
)

API_RESPONSE_PROMPT_TEMPLATE = (
    API_URL_PROMPT_TEMPLATE
    + """ {api_url}

以下是API的响应结果：

{api_response}

你应该一步一步思考：
1. 你从API的响应结果看到了什么，有哪些内容和问题有关系
2. 进一步分析该响应结果能否回答问题
3. 用科学工作者严谨的口吻来组织你的语言去回答用户提出的这个问题，你需要提供问题的答案和依据
4. 请勿编造你的回答

你对问题的回复:"""
)

API_RESPONSE_PROMPT = PromptTemplate(
    input_variables=["api_docs", "question", "api_url", "api_response"],
    template=API_RESPONSE_PROMPT_TEMPLATE,
)



import json
from typing import Any, Dict

from pydantic import BaseModel, Extra, root_validator


class ArxivAPIWrapper(BaseModel):
    """Wrapper around ArvixAPI.

    To use, you should have the ``arxiv`` python package installed.
    This wrapper will use the Arxiv API to conduct searches and
    fetch page summaries. By default, it will return the page summaries
    of the top-k results of an input search.
    """
    arxiv_client: Any
    top_k_results: int = 3

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.ignore

    @root_validator()
    def validate_environment(cls, values: Dict) -> Dict:
        """Validate that the python package exists in environment."""
        try:
            import arxiv

            values["arxiv_client"] = arxiv.Client(
                page_size=1,
                delay_seconds=3,  # "arXiv's Terms of Use" asks you no more than one request per three seconds
                num_retries=3
            )
        except ImportError:
            raise ValueError(
                "Could not import arxiv python package. "
                "Please it install it with `pip install arxiv`."
            )
        return values

    def run(self, query_json_str: str) -> str:
        import arxiv

        query_json = json.loads(query_json_str)

        search_query = query_json.get("search_query", "")
        max_results = query_json.get("max_results", self.top_k_results)
        sort_by = query_json.get("sort_by", "relevance")
        sort_order = query_json.get("sort_order", "descending")

        # todo avoid it in future
        query = search_query.replace("'", "").replace("\"", "").replace("+", " ")

        if sort_by == "lastUpdatedDate":
            sort_by_input = arxiv.SortCriterion.LastUpdatedDate
        elif sort_by == "submittedDate":
            sort_by_input = arxiv.SortCriterion.SubmittedDate
        else:
            sort_by_input = arxiv.SortCriterion.Relevance

        if sort_order == "ascending":
            sort_order_input = arxiv.SortOrder.Ascending
        else:
            sort_order_input = arxiv.SortOrder.Descending

        search = arxiv.Search(query=query,
                              max_results=max_results,
                              sort_by=sort_by_input,
                              sort_order=sort_order_input)

        _content = ""
        for idx, result in enumerate(self.arxiv_client.results(search)):
            # 标题、作者、published、总结、primary_category、comment、pdf_url
            _content += f"\n第{idx+1}篇：{repr(result.title)}\n"
            _content += f"作者: {[author.name for author in result.authors]}\n"
            try:
                _content += f"发布时间: {result.published.strftime('%Y-%m-%d %H:%M:%S')}\n"
            except Exception as e:
                _content += f"发布时间: None\n"
            _content += f"摘要: {result.summary}\n"
            _content += f"分类: {result.primary_category}\n"
            if result.comment:
                _content += f"备注: {repr(result.comment)}\n"
            _content += f"pdf: {result.pdf_url}\n"
            _content += "\n---\n"
        return _content



import logging
from typing import Any

from rich.console import Console

from chatgpt_tool_hub.common.log import LOG
from chatgpt_tool_hub.tools.all_tool_list import main_tool_register
from chatgpt_tool_hub.tools.base_tool import BaseTool
from chatgpt_tool_hub.tools.web_requests import BaseRequestsTool, filter_text, RequestsWrapper

default_tool_name = "url-get"


class RequestsGetTool(BaseRequestsTool, BaseTool):
    """Tool for making a GET request to an API endpoint."""

    name = default_tool_name
    description = (
        "A portal to the internet. Use this when you need to get specific content from a website. "
        "Input should be a url (i.e. https://www.google.com). "
        "The output will be the text response of the GET request."
    )

    def __init__(self, console: Console = Console(), **tool_kwargs: Any):
        # 这个工具直接返回内容
        super().__init__(console=console, requests_wrapper=RequestsWrapper(**tool_kwargs), return_direct=False)

    def _run(self, url: str) -> str:
        """Run the tool."""
        try:
            html = self.requests_wrapper.get(url)
            _content = filter_text(html)
            LOG.debug(f"[url-get] output: {str(_content)}")
        except Exception as e:
            LOG.error(f"[url-get] {str(e)}")
            _content = repr(e)
        return _content

    async def _arun(self, url: str) -> str:
        """Run the tool asynchronously."""
        try:
            html = await self.requests_wrapper.aget(url)
            _content = filter_text(html)
            LOG.debug(f"[url-get] output: {str(_content)}")
        except Exception as e:
            LOG.error(f"[url-get] {str(e)}")
            _content = repr(e)
        return _content


main_tool_register.register_tool(default_tool_name, lambda console, kwargs: RequestsGetTool(console, **kwargs), [])


if __name__ == "__main__":
    LOG.setLevel(logging.DEBUG)
    requests_wrapper = RequestsWrapper()
    tool = RequestsGetTool(requests_wrapper=requests_wrapper)
    content = tool.run("https://github.com/goldfishh/chatgpt-tool-hub")
    print(content)



"""An bot designed to hold a conversation in addition to using tools."""
from __future__ import annotations

import json
import traceback
from typing import Any, List, Optional, Sequence, Tuple, Union, Dict

from rich.console import Console
from rich.panel import Panel
from chatgpt_tool_hub.bots.chat_bot.prompt import FORMAT_INSTRUCTIONS, PREFIX, SUFFIX
from chatgpt_tool_hub.chains import LLMChain
from chatgpt_tool_hub.common import json_utils
from chatgpt_tool_hub.common.callbacks import BaseCallbackManager
from chatgpt_tool_hub.common.log import LOG
from chatgpt_tool_hub.common.schema import BotAction, BotFinish
from chatgpt_tool_hub.engine import Bot
from chatgpt_tool_hub.models.base import BaseLLM
from chatgpt_tool_hub.prompts import PromptTemplate
from chatgpt_tool_hub.tools.base_tool import BaseTool

default_ai_prefix = "LLM-OS"
default_human_prefix = "user"


class ChatBot(Bot):
    """An bot designed to hold a conversation in addition to using tools."""

    ai_prefix: str = default_ai_prefix
    human_prefix: str = default_human_prefix

    @property
    def _bot_type(self) -> str:
        """Return Identifier of bot type."""
        return "chat-bot"

    @property
    def observation_prefix(self) -> str:
        """Prefix to append the observation with."""
        return "Observation: "

    def _fix_text(self, text: str) -> str:
        tool_names = ", ".join(list(self.allowed_tools)) if self.allowed_tools else ""
        instruction_text = FORMAT_INSTRUCTIONS.format(
            tool_names=tool_names, human_prefix=self.human_prefix
        )
        return f"\n\nYou just told me: {text}, but it doesn't meet the format I mentioned to you. \n\nformat: {instruction_text}. \n\nYou should understand why you did not input the correct format, correct it and try again. \n\n"

    @property
    def llm_prefix(self) -> str:
        """Prefix to append the llm call with."""
        return "Thought:"

    @classmethod
    def create_prompt(
            cls,
            tools: Sequence[BaseTool],
            prefix: str = PREFIX,
            suffix: str = SUFFIX,
            format_instructions: str = FORMAT_INSTRUCTIONS,
            ai_prefix: str = default_ai_prefix,
            human_prefix: str = default_human_prefix,
            input_variables: Optional[List[str]] = None,
    ) -> PromptTemplate:
        """Create prompt in the style of the zero shot bot.

        Args:
            tools: List of tools the bot will have access to, used to format the
                prompt.
            prefix: String to put before the list of tools.
            suffix: String to put after the list of tools.
            ai_prefix: String to use before AI output.
            human_prefix: String to use before human output.
            input_variables: List of input variables the final prompt will expect.

        Returns:
            A PromptTemplate with the template assembled from the pieces here.
        """
        tool_strings = "\n".join(
            [f"> {tool.name}: {tool.description}" for tool in tools]
        )
        tool_names = ", ".join([tool.name for tool in tools])

        instruction_text = format_instructions.format(human_prefix=human_prefix, tool_names=tool_names)
        prefix = prefix.format(human_prefix=human_prefix)

        template = "\n\n".join([prefix, tool_strings, instruction_text, suffix])
        if input_variables is None:
            input_variables = ["input", "chat_history", "bot_scratchpad"]
        return PromptTemplate(template=template, input_variables=input_variables)

    def plan(
        self, intermediate_steps: List[Tuple[BotAction, str]], **kwargs: Any
    ) -> Union[BotAction, BotFinish]:
        """Given input, decided what to do."""
        full_inputs = self.get_full_inputs(intermediate_steps, **kwargs)
        action = self._get_next_action(full_inputs)
        if action.tool == "answer-user":
            return BotFinish({"output": action.tool_input}, action.log)
        return action

    def get_full_inputs(
        self, intermediate_steps: List[Tuple[BotAction, str]], **kwargs: Any
    ) -> Dict[str, Any]:
        """Create the full inputs for the LLMChain from intermediate steps."""
        thoughts = self._construct_scratchpad(intermediate_steps)
        # todo remove stop
        new_inputs = {"bot_scratchpad": self._crop_full_input(thoughts), "stop": self._stop}
        return {**kwargs, **new_inputs}

    def _get_next_action(self, full_inputs: Dict[str, str]) -> BotAction:
        llm_answer_str = self.llm_chain.predict(**full_inputs)

        action, action_input = self._extract_tool_and_input(llm_answer_str)

        LOG.info(f"输入: {action_input}")
        
        # json 解析错误重试
        retry_num = 0
        while not action:
            retry_num += 1
            if retry_num > self.max_parse_retry_num:
                raise ValueError(f"Could not parse LLM output: `{llm_answer_str}`")

            full_output = self._fix_text(llm_answer_str)
            full_inputs["bot_scratchpad"] += full_output
            output = self.llm_chain.predict(**full_inputs)

            action, action_input = self._extract_tool_and_input(output)

            LOG.info(f"重试输入: {action_input}")
        return BotAction(
            tool=action, tool_input=action_input, log=llm_answer_str
        )

    @property
    def finish_tool_name(self) -> str:
        """Name of the tool to use to finish the chain."""
        return self.ai_prefix

    def _extract_tool_and_input(self, llm_output: str) -> Optional[Tuple[str, str]]:
        # 1. json loads and 修复 & 错误处理
        llm_reply_json = self.parse_reply_json(llm_output)
        action, action_input = "", ""

        try:
            if not isinstance(llm_reply_json, dict):
                return "Error:", f"'response_json' object is not dictionary {llm_reply_json}"

            if "tool" not in llm_reply_json:
                return "Error:", "Missing 'tool' object in JSON"

            tool_dict = llm_reply_json["tool"]
            if not isinstance(tool_dict, dict):
                return "Error:", "'tool' object is not a dictionary"

            if "name" not in tool_dict:
                return "Error:", "Missing 'name' field in 'tool_dict' object"

            action = tool_dict["name"]
            action_input = tool_dict.get("input", "")

        except json.decoder.JSONDecodeError:
            LOG.error("Error:", "Invalid JSON")

        except Exception as e:
            LOG.error(f"Error: {repr(e)}")

        if action.lower() == "answer-user" and action_input in ['', '空', '无']:
            action_input = llm_reply_json.get('thoughts', {}).get('speak', '')

        if action.lower() not in self.allowed_tools:
            return "answer-user", action_input

        if action.lower() != "answer-user":
            self.console.print(f"√ 我在用 [bold cyan]{action}[/] 工具...")

        LOG.info(f"执行Tool: {action}中...")
        return action.strip(), action_input.strip()

    def parse_reply_json(self, assistant_reply) -> dict:
        """Prints the assistant's thoughts to the console"""
        try:
            try:
                # Parse and print Assistant response
                assistant_reply_json = json_utils.fix_and_parse_json(assistant_reply)
            except json.JSONDecodeError as e:
                LOG.error(f"Error: Invalid JSON in assistant thoughts: {assistant_reply}, error: {repr(e)}")
                assistant_reply_json = json_utils.fix_json_by_finding_outermost_brackets(assistant_reply)

                assistant_reply_json = json_utils.fix_and_parse_json(assistant_reply_json)

            # Check if assistant_reply_json is a string and attempt to parse it into a
            #  JSON object
            if isinstance(assistant_reply_json, str):
                try:
                    assistant_reply_json = json.loads(assistant_reply_json)
                except json.JSONDecodeError as e:
                    LOG.error(f"Error: Invalid JSON: {assistant_reply}, error: {repr(e)}")
                    assistant_reply_json = (
                        json_utils.fix_json_by_finding_outermost_brackets(assistant_reply_json)
                    )

            assistant_thoughts = assistant_reply_json.get("thoughts", {})
            assistant_thoughts_text = assistant_thoughts.get("text")
            assistant_thoughts_reasoning = None
            assistant_thoughts_criticism = None
            assistant_thoughts_speak = None

            if assistant_thoughts:
                assistant_thoughts_reasoning = assistant_thoughts.get("reasoning")
                assistant_thoughts_criticism = assistant_thoughts.get("criticism")
                assistant_thoughts_speak = assistant_thoughts.get("speak")

            thoughts_text = ""
            if assistant_thoughts_reasoning:
                thoughts_text += f"推理：{assistant_thoughts_reasoning}\n"
            if assistant_thoughts_text:
                thoughts_text += f"思考：{assistant_thoughts_text}\n"
            if assistant_thoughts_criticism:
                thoughts_text += f"反思：{assistant_thoughts_criticism}\n"

            self.console.print(Panel(thoughts_text,
                                     title=f"{self.ai_prefix.upper()}的内心独白",
                                     highlight=True, style='dim'))
            # it's useful for avoid splitting Panel
            LOG.info(f"{self.ai_prefix.upper()}的内心独白: {thoughts_text}")
            
            return assistant_reply_json
        except json.decoder.JSONDecodeError as e:
            call_stack = traceback.format_exc()
            LOG.error(f"Error: Invalid JSON: {assistant_reply}\n")
            LOG.error(f"Traceback: {repr(call_stack)}, error: {repr(e)}")

        # All other errors, return "Error: + error message"
        except Exception as e:
            call_stack = traceback.format_exc()
            LOG.error(f"Traceback: {repr(call_stack)}, error: {repr(e)}")

    @classmethod
    def from_llm_and_tools(
            cls,
            llm: BaseLLM,
            tools: Sequence[BaseTool],
            console: Console = Console(),
            callback_manager: Optional[BaseCallbackManager] = None,
            prefix: str = PREFIX,
            suffix: str = SUFFIX,
            format_instructions: str = FORMAT_INSTRUCTIONS,
            ai_prefix: str = default_ai_prefix,
            human_prefix: str = default_human_prefix,
            input_variables: Optional[List[str]] = None,
            **kwargs: Any,
    ) -> Bot:
        """Construct an bot from an LLM and tools."""
        prompt = cls.create_prompt(
            tools,
            ai_prefix=ai_prefix,
            human_prefix=human_prefix,
            prefix=prefix,
            suffix=suffix,
            format_instructions=format_instructions,
            input_variables=input_variables,
        )
        llm_chain = LLMChain(
            llm=llm,
            prompt=prompt,
            callback_manager=callback_manager,
        )
        tool_names = [tool.name for tool in tools]
        return cls(
            llm_chain=llm_chain, console=console, allowed_tools=tool_names, ai_prefix=ai_prefix, **kwargs
        )



from chatgpt_tool_hub.prompts import PromptTemplate

PROMPT = """
你是文案编辑，你的工作是组合下列的文本，同时做一些转述

{text} 

组合要求：
1. 组合时必须是中文
2. 组合时必须保留所有内容细节和前后关系
3. 需要你用容易理解的方式转述
4. 不低于120字，不超过800字
5. 你不能捏造信息

Output:
"""

REDUCE_PROMPT = PromptTemplate(
    input_variables=["text"],
    template=PROMPT,
)



from typing import Any

from rich.console import Console

from chatgpt_tool_hub.chains.api import APIChain
from chatgpt_tool_hub.common.utils import get_from_dict_or_env
from chatgpt_tool_hub.models import build_model_params
from chatgpt_tool_hub.models.model_factory import ModelFactory
from chatgpt_tool_hub.tools.base_tool import BaseTool
from chatgpt_tool_hub.tools.news import news_tool_register
from chatgpt_tool_hub.tools.news.news_api.docs_prompt import NEWS_DOCS

default_tool_name = "news-api"


class NewsApiTool(BaseTool):
    name: str = default_tool_name
    description: str = (
        "Use this when you want to get information about the top headlines of current news stories. "
        "The input should be a question in natural language that this API can answer."
    )
    api_chain: APIChain = None

    def __init__(self, console: Console = Console(), **tool_kwargs: Any):
        super().__init__(console=console)

        news_api_key = get_from_dict_or_env(tool_kwargs, "news_api_key", "NEWS_API_KEY")

        llm = ModelFactory().create_llm_model(**build_model_params(tool_kwargs))

        self.api_chain = APIChain.from_llm_and_api_docs(
            llm, NEWS_DOCS, console=console, headers={"X-Api-Key": news_api_key}
        )

    def _run(self, query: str) -> str:
        """Use the tool."""
        if not query:
            return "the input of tool is empty"
        return (
            self.api_chain.run(query)
            if self.api_chain
            else "the tool was not initialized"
        )

    async def _arun(self, query: str) -> str:
        """Use the tool asynchronously."""
        raise NotImplementedError("NewsTool does not support async")


news_tool_register.register_tool(default_tool_name, lambda console, kwargs: NewsApiTool(console, **kwargs), ["news_api_key"])



from datetime import datetime

from rich.console import Console

from chatgpt_tool_hub.chains.api import APIChain
from chatgpt_tool_hub.models import build_model_params
from chatgpt_tool_hub.models.model_factory import ModelFactory
from chatgpt_tool_hub.tools.all_tool_list import main_tool_register
from chatgpt_tool_hub.tools.base_tool import BaseTool
from chatgpt_tool_hub.tools.meteo.docs_prompt import OPEN_METEO_DOCS

default_tool_name = "meteo-weather"


class MeteoWeatherTool(BaseTool):
    name: str = default_tool_name
    description: str = (
        "当你想要获取天气信息时应该使用本工具。"
        "你需要分析用户想要哪些天气信息，严谨地用自然语言描述问题，然后将该问题传入本工具。"
        "你最好能在输入末尾添加返回天气信息的粒度，支持：按时 或 按天，查询某天的天气优先用按时，遇到错误可以更换粒度"
    )
    api_chain: APIChain = None

    def __init__(self, console: Console = Console(), **tool_kwargs):
        super().__init__(console=console, return_direct=False)
        llm = ModelFactory().create_llm_model(**build_model_params(tool_kwargs))
        self.api_chain = APIChain.from_llm_and_api_docs(llm, OPEN_METEO_DOCS, console=console)

    def _run(self, query: str) -> str:
        """Use the tool."""
        if not query:
            return "the input of tool is empty"
        if not self.api_chain:
            return "the tool was not initialized"

        query += f"\nThe current time is {datetime.now().strftime('%Y-%m-%d %H:%M:%S')} in UTC+8."

        return self.api_chain.run(query)

    async def _arun(self, query: str) -> str:
        """Use the tool asynchronously."""
        raise NotImplementedError("MeteoWeatherTool does not support async")


main_tool_register.register_tool(default_tool_name, lambda console, kwargs: MeteoWeatherTool(console, **kwargs), [])



from typing import List

from chatgpt_tool_hub.common.log import LOG


class ToolRegister:
    def __init__(self):
        self.REGISTER_TOOLS = {}

    def register_tool(self, name: str, function: callable, tool_input_keys: list):
        self.REGISTER_TOOLS[name] = (function, tool_input_keys)

    def unregister_tool(self, name: str) -> bool:
        if self.REGISTER_TOOLS.get(name):
            self.REGISTER_TOOLS.pop(name)
            return True
        else:
            LOG.warning(f"unregister a unknown tool: {name}")
            return False

    def get_registered_tool_names(self) -> List[str]:
        """Get a list of all possible tool names."""
        return list(self.REGISTER_TOOLS)

    def get_registered_tool(self) -> dict:
        return self.REGISTER_TOOLS



"""Util that calls Bing Search."""
import json
from typing import Dict, List

from pydantic import BaseModel, Extra, validator, root_validator

from chatgpt_tool_hub.common.log import LOG
from chatgpt_tool_hub.common.utils import get_from_dict_or_env
from chatgpt_tool_hub.tools.web_requests import filter_text
from chatgpt_tool_hub.tools.web_requests.wrapper import RequestsWrapper


class BingSearchAPIWrapper(BaseModel):
    """Wrapper for Bing Search API."""

    bing_subscription_key: str
    bing_search_url: str
    top_k_results: int = 2

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.ignore

    def _bing_search_results(self, search_term: str, count: int) -> List[dict]:
        headers = {"Ocp-Apim-Subscription-Key": self.bing_subscription_key}
        params = {
            "q": search_term,
            "count": count,
            "textDecorations": True,
            "textFormat": "HTML",
        }
        response = RequestsWrapper(headers=headers).get(self.bing_search_url, params=params, raise_for_status=True)

        search_results = json.loads(response)
        try:
            result = search_results["webPages"]["value"]
            LOG.debug(f"[bing_search] output: {str(result)}")
        except Exception as e:
            result = []
            LOG.error(f"[bing_search] {repr(e)}")
        return result

    @root_validator(pre=True)
    def validate_environment(cls, values: Dict) -> Dict:
        """Validate that api key and endpoint exists in environment."""

        values["bing_subscription_key"] = get_from_dict_or_env(
            values, "bing_subscription_key", "BING_SUBSCRIPTION_KEY"
        )

        values["bing_search_url"] = get_from_dict_or_env(
            values,
            "bing_search_url",
            "BING_SEARCH_URL",
            default="https://api.bing.microsoft.com/v7.0/search",
        )

        values["top_k_results"] = get_from_dict_or_env(
            values, 'top_k_results', "TOP_K_RESULTS", 2
        )

        return values

    def run(self, query: str) -> str:
        """Run query through BingSearch and parse result."""
        results = self._bing_search_results(query, count=self.top_k_results)
        if len(results) == 0:
            return "No good Bing Search Result was found"

        snippets = [filter_text(result["snippet"]) for result in results]
        return " ".join(snippets)

    def results(self, query: str, num_results: int) -> List[Dict]:
        """(for json result): Run query through BingSearch and return metadata.

        Args:
            query: The query to search for.
            num_results: The number of results to return.

        Returns:
            A list of dictionaries with the following keys:
                snippet - The description of the result.
                title - The title of the result.
                link - The link to the result.
        """
        metadata_results = []
        results = self._bing_search_results(query, count=num_results)
        if len(results) == 0:
            return [{"Result": "No good Bing Search Result was found"}]
        for result in results:
            metadata_result = {
                "snippet": result["snippet"],
                "title": result["name"],
                "link": result["url"],
            }
            metadata_results.append(metadata_result)

        return metadata_results



from typing import List

from chatgpt_tool_hub.tools.tool_register import ToolRegister


main_tool_register = ToolRegister()


# used by chatgpt-on-wechat now , don't move it
def get_all_tool_names() -> List[str]:
    """Get a list of all possible tool names."""
    return main_tool_register.get_registered_tool_names()
